#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include "main.h"
#include "socket.h"
#include "data.h"
#include "commands.h"

#define DEBUG 		0
#include "debug.h"

#define VERSION				"0.1"
#define PROTOCOL_VERSION	"1"

#define IDENTIFY			"SERVER " PROTOCOL_VERSION " \"Netboard server "VERSION"\"\n"
#define HEADER				"NetBoard server " VERSION " ("PROTOCOL_VERSION")\n"		\
							"Josef Kufner <jk@myserver.cz>\n\n"

#define BUF_SIZE	512

static const char *server_id = IDENTIFY;

struct s_cfg cfg = {
	3003,		/* port */
	NULL,		/* host */
	128,		/* max_connections */
	64,			/* max games */
	0,			/* verbose */
};

int server_fd = 0;
CLIENT *client = NULL;

void *malloc_c(size_t size) {
	void *p;
	
	if ((p = malloc(size)) != NULL) {
		return p;
	} else {
		perror("netboard-server");
		exit(127);
	}
}

void on_signal(int n) {
	if (n == SIGINT) {
		exit(0);
	}
}

void cleanup_on_exit(void) {
	int i;

	if (cfg.verbose) {
			printf("Exiting...\n");
	}
	
	for (i = 0; client && i < cfg.max_connections; i++) {
		if (client[i].fd)
			leave_client(&client[i]);
	}
	close_socket(server_fd);
}

int main (int argc, char *argv[]) {
	int opt, x, i, new_client, bytes;
	fd_set read_fds;
	char buf[BUF_SIZE];

	signal(SIGINT, on_signal);

	/* Args */
	while ((opt = getopt(argc, argv, "g:l:mp:vh")) != -1) {
		switch(opt) {
			case 'g':
				if(sscanf(optarg, "%d", &x) == 1) {
					cfg.max_games = x;
				}
				break;
			case 'l':
				if (cfg.host)
					free(cfg.host);
				cfg.host = strdup(optarg);
				break;
			case 'm':
				if(sscanf(optarg, "%d", &x) == 1) {
					cfg.max_connections = x;
				}
				break;
			case 'p':
				if(sscanf(optarg, "%u", &x) == 1) {
					cfg.port = x;
				}
				break;
			case 'v':
				cfg.verbose++;
				break;
			case 'h':
				printf(HEADER
						"Options:\n"
						"\t-g number    sets max games\n"
						"\t-m number    sets max connections\n"
						"\t-p port      sets tcp port to listen\n"
						"\t-v           be verbose\n"
						"\t-h           this help\n"
						"\n");
				return 0;
		}
	}

	atexit(cleanup_on_exit);

	/* default host */
	if (!cfg.host) {
		cfg.host = "0.0.0.0";
	}

	/* header if verbose */
	if (cfg.verbose) {
		printf(HEADER		
			"tcp port:         %u\n"
			"max connections:  %d\n"
			"\n", cfg.port, cfg.max_connections);
	}

	/* init socket */
	if ((server_fd = create_socket(cfg.host, cfg.port)) == -1) {
		exit(-1);
	}

	/* create array for clients */
	client = create_clients(cfg.max_connections);
	
	/* create array for games */
	game_list = create_game_list(cfg.max_games);
	
	/* main loop */
	while (1) {
		FD_ZERO(&read_fds);
		FD_SET(server_fd, &read_fds);
		for (i = 0; i < cfg.max_connections; i++) {
			if (client[i].fd)
				FD_SET(client[i].fd, &read_fds);
		}

		switch(select(FD_SETSIZE, &read_fds, NULL, NULL, NULL)) {
			case 0:  /* timeout */
				break;
				
			case -1: /* error */
				perror("netboard-server: select");
				return -1;

			default: /* fd activity */
				if (FD_ISSET(server_fd, &read_fds)) {
					/* new connection*/
					new_client = accept_client(server_fd);
					for (i = 0; i < cfg.max_connections; i++) {
						if (client[i].fd == 0) {
							client[i].fd = new_client;
							write(client[i].fd, server_id, strlen(server_id));
							if (cfg.verbose) {
								printf("New connection (socket %d)\n", client[i].fd);
							}
							break;
						}
					}
					if (i == cfg.max_connections) {
						close_socket(new_client);
						if (cfg.verbose) {
							printf("New connection rejected (socket %d)\n", new_client);
						}
					}
				} else {
					/* client activity */
					for (i = 0; i < cfg.max_connections; i++) {
						if (FD_ISSET(client[i].fd, &read_fds)) {
							bytes = read(client[i].fd, buf, BUF_SIZE - 1);
							
							if (bytes > 0) {
								/* data */
								buf[bytes] = '\0';
								DBG_MSG(6, "Received from socket %d: %s", client[i].fd, buf);
								
								if (parse_cmd(&client[i], buf)) {
									if (client[i].fd) {
										DBG_MSG(8, "main loop: Sending OK to socket %d...", client[i].fd);
										write(client[i].fd, "OK\n", 3);
										DBG_MSG(10, "main loop: OK sent");
									}
								} else {
									if (client[i].fd) {
										DBG_MSG(8, "main loop: Sending ERR to socket %d...", client[i].fd);
										write(client[i].fd, "ERR\n", 4);
										DBG_MSG(10, "main loop: Sent ERR");
									}
								}
							} else {
								/* socked closed */
								if (cfg.verbose) {
									printf("Client \"%s\" (socket %d) has left.\n",
											client[i].name ? client[i].name : "", client[i].fd);
								}
								client[i].fd = 0;
								leave_client(&client[i]);
							}
						}
					}
				}
				break;
		}
	}

	return 0;
}


/* vim:set ts=4 sw=4: */
